##################### generated by xml-casa (v2) from msuvbinflag.xml ###############
##################### ee56599cfec6ed7433808e6b2bd337c8 ##############################
from __future__ import absolute_import
from casashell.private.stack_manip import find_local as __sf__
from casashell.private.stack_manip import find_frame as _find_frame
from casatools.typecheck import validator as _pc
from casatools.coercetype import coerce as _coerce
from casatools.errors import create_error_string
from casatasks.private.task_msuvbinflag import msuvbinflag as _msuvbinflag_t
from casatasks.private.task_logging import start_log as _start_log
from casatasks.private.task_logging import end_log as _end_log
from collections import OrderedDict
import numpy
import sys
import os

import shutil

def static_var(varname, value):
    def decorate(func):
        setattr(func, varname, value)
        return func
    return decorate

class _msuvbinflag:
    """
    msuvbinflag ---- An algorithm to identify outliers in the UV plane through defined uniform grids by msuvbin.
    Both msuvbin and msuvbinflag tasks are experimental
    

    
    Msuvbinflag is an automatic flagging algorithm for the identification of Radio Frequency Interference (RFI)
    in the UV plane for experimental purpose. The visibilities in input CASA Measurements Set (MS) is
    first gridded into an uniformed UV plane with msuvbin task.
    Then the significant outliers are computed for entire UV grid based on the settings configurable by users for
    three method options: radian, regional mean and gradient. After computing flags for each visibility in each
    uniformed UV bin, the visibilities are transformed back to original order in the MS file with flagging information.
    The first msuvbinflag release version is for NRAO internal users to evaluate the performance for both
    computational and numerical analysis.
    
    

    --------- parameter descriptions ---------------------------------------------

    binnedvis  Name of input uniform binned visibility data (MS) set on UV plane using msuvbin task.
               default: none;
               example: binnedvis='SNR_Binnedvis.ms'
    method     There are four algorithm options for selecting the outliers as flags (radial. regionalMean, peak,gradient)
               method='radial':  This algorithm is converted the uv in cartesian coordinates into polarization coordinates along with
                                 frequency and weighting data. Then average values are calculated for each annular circle with one unit equal spacing.
                                 A binary linear least square fitting is generated followed by the comparison analysis between the original value and the
                                 fitting value. A point is treated as an outliers to be flagged if original value is greater than the sum of
                                 the fitting value and the multiple of the user defined sigma and the interpolated value.
                                 eg: originalValue[x,y] > radiafitValue[r]+sigma*interpolateValue[r].
               
               method='regionalMean': This algorithm is to compute the statistical mean, median and standard deviation for every point
                                      as a center within a square region which the size is defined by user. A point is treated as an
                                      outliers to be flagged if it is beyond the certain criteria threshold.
                                      eg: originalValue[x,y]) > regionalMean+sigma*regionalStd
               
               method='median': This algorithm is to flag according to median criteria.
               
               method='gradient': This algorithm is to compute the gradient for every point. It is usually called after the peak falgging algorithm.
                                   Then a point is treated as an outliers to be flagged if the gradient is greater than the user
                                  defined sigma. eg: gradient > sigmaGradient. After running through second time msuvbin, another comparison between
                                   user defined size of square regional median to total median, a point is treated as an outliers to be flagged if
                                   it is beyond the certain criteria threshold. eg: originalValue[x,y] > sigma * regionalStd[x,y]*median/regionalMedian.
               
               default: radial
               example: method='radial'
    sigma      User defined sigma value for comparison for radia algorithm. It is double type.
               The default value is 5.
    sizeRegion User defined local square regional size for "regionalMean" and "gradient" algorithm option.
               It is more like the size of a moving window in Numpy 2D convolution.
    radius     User defined radius value for gradient algorithm. It is double type.
               The default value is 1.
    ignorPoint User defined scale value for gradient comparison for gradient algorithm. It is bool type.
               The default value is False.
    [1;42mRETURNS[1;m       void

    --------- examples -----------------------------------------------------------

    
    
    
    msuvbinflag(binnedvis='SNR_G55_10s.ms', improveImp=True, method='radial', sigma=5)
    
    Keyword arguments:
    binnedvis -- Name of input binned visibility file which is output from msuvbin.
    After msuvbinflag, binnedvis is modified with flagging info.
    
    method -- Three method options: "radian", "regional mean","median","gradient"
    default: 'radian'
    
    sigmaRadian -- User set sigma value for radian algorithm.
    default=5.0
    
    


    """

    _info_group_ = """manipulation, imaging"""
    _info_desc_ = """An algorithm to identify outliers in the UV plane through defined uniform grids by msuvbin.
      Both msuvbin and msuvbinflag tasks are experimental
  """

    __schema = {'binnedvis': {'type': 'cStr', 'coerce': _coerce.to_str}, 'method': {'type': 'cVariant', 'coerce': [_coerce.to_variant] # <allowed> IS NOT ALLOWED FOR A PARAMETER OF TYPE any
}, 'sigma': {'type': 'cFloat', 'coerce': _coerce.to_float}, 'sizeRegion': {'type': 'cFloat', 'coerce': _coerce.to_float}, 'radius': {'type': 'cFloat', 'coerce': _coerce.to_float}, 'ignorPoint': {'type': 'cBool'}}

    def __init__(self):
        self.__stdout = None
        self.__stderr = None
        self.__root_frame_ = None

    def __globals_(self):
        if self.__root_frame_ is None:
            self.__root_frame_ = _find_frame( )
            assert self.__root_frame_ is not None, "could not find CASAshell global frame"
        return self.__root_frame_

    def __to_string_(self,value):
        if type(value) is str:
            return "'%s'" % value
        else:
            return str(value)

    def __validate_(self,doc,schema):
        return _pc.validate(doc,schema)

    def __do_inp_output(self,param_prefix,description_str,formatting_chars):
        out = self.__stdout or sys.stdout
        description = description_str.split( )
        prefix_width = 23 + 13 + 4
        output = [ ]
        addon = ''
        first_addon = True
        if len(description) == 0:
            out.write(param_prefix + " #\n")
            return
        while len(description) > 0:
            ## starting a new line.....................................................................
            if len(output) == 0:
                ## for first line add parameter information............................................
                if len(param_prefix)-formatting_chars > prefix_width - 1:
                    output.append(param_prefix)
                    continue
                addon = param_prefix + ' #'
                first_addon = True
                addon_formatting = formatting_chars
            else:
                ## for subsequent lines space over prefix width........................................
                addon = (' ' * prefix_width) + '#'
                first_addon = False
                addon_formatting = 0
            ## if first word of description puts us over the screen width, bail........................
            if len(addon + description[0]) - addon_formatting + 1 > self.term_width:
                ## if we're doing the first line make sure it's output.................................
                if first_addon: output.append(addon)
                break
            while len(description) > 0:
                ## if the next description word puts us over break for the next line...................
                if len(addon + description[0]) - addon_formatting + 1 > self.term_width: break
                addon = addon + ' ' + description[0]
                description.pop(0)
            output.append(addon)
        out.write('\n'.join(output) + '\n')

    #--------- return nonsubparam values ----------------------------------------------

    def __method_dflt( self, glb ):
        return 'radial'

    def __method( self, glb ):
        if 'method' in glb: return glb['method']
        return 'radial'

    def __binnedvis_dflt( self, glb ):
        return ''

    def __binnedvis( self, glb ):
        if 'binnedvis' in glb: return glb['binnedvis']
        return ''



    #--------- return inp/go default --------------------------------------------------
    def __sizeRegion_dflt( self, glb ):
        if self.__method( glb ) == "regionalMean": return float(20)
        return None
    def __ignorPoint_dflt( self, glb ):
        if self.__method( glb ) == "regionalMean": return bool(True)
        return None
    def __sigma_dflt( self, glb ):
        if self.__method( glb ) == "median": return float(0.3)
        if self.__method( glb ) == "gradient": return float(0.3)
        return None
    def __radius_dflt( self, glb ):
        if self.__method( glb ) == "gradient": return float(10)
        return None

    #--------- return subparam values -------------------------------------------------
    def __sigma( self, glb ):
        if 'sigma' in glb: return glb['sigma']
        dflt = self.__sigma_dflt( glb )
        if dflt is not None: return dflt
        return float(5)
    def __sizeRegion( self, glb ):
        if 'sizeRegion' in glb: return glb['sizeRegion']
        dflt = self.__sizeRegion_dflt( glb )
        if dflt is not None: return dflt
        return float(20)
    def __radius( self, glb ):
        if 'radius' in glb: return glb['radius']
        dflt = self.__radius_dflt( glb )
        if dflt is not None: return dflt
        return float(1)
    def __ignorPoint( self, glb ):
        if 'ignorPoint' in glb: return glb['ignorPoint']
        dflt = self.__ignorPoint_dflt( glb )
        if dflt is not None: return dflt
        return True

    #--------- subparam inp output ----------------------------------------------------
    def __binnedvis_inp(self):
        def xml_default( ):
            ## play the crazy subparameter shell game
            dflt = self.__binnedvis_dflt( self.__globals_( ) )
            if dflt is not None: return dflt
            return ''
        description = 'Name of input binned MS from msuvbin task.'
        value = self.__binnedvis( self.__globals_( ) )
        (pre,post) = (('','') if value == xml_default( ) else ('\x1B[34m','\x1B[0m')) if self.__validate_({'binnedvis': value},{'binnedvis': self.__schema['binnedvis']}) else ('\x1B[91m','\x1B[0m')
        self.__do_inp_output('%-13.13s = %s%-23s%s' % ('binnedvis',pre,self.__to_string_(value),post),description,0+len(pre)+len(post))
    def __method_inp(self):
        def xml_default( ):
            ## play the crazy subparameter shell game
            dflt = self.__method_dflt( self.__globals_( ) )
            if dflt is not None: return dflt
            return 'radial'
        description = 'Flagging method option.'
        value = self.__method( self.__globals_( ) )
        (pre,post) = (('','') if value == xml_default( ) else ('\x1B[34m','\x1B[0m')) if self.__validate_({'method': value},{'method': self.__schema['method']}) else ('\x1B[91m','\x1B[0m')
        self.__do_inp_output('\x1B[1m\x1B[47m%-13.13s =\x1B[0m %s%-23s%s' % ('method',pre,self.__to_string_(value),post),description,13+len(pre)+len(post))
    def __sigma_inp(self):
        def xml_default( ):
            ## play the crazy subparameter shell game
            dflt = self.__sigma_dflt( self.__globals_( ) )
            if dflt is not None: return dflt
            return float(5)
        if self.__sigma_dflt( self.__globals_( ) ) is not None:
             description = 'User defined sigma value for comparison for radial algorithm and gradient algorithm.'
             value = self.__sigma( self.__globals_( ) )
             (pre,post) = (('','') if value == xml_default( ) else ('\x1B[34m','\x1B[0m')) if self.__validate_({'sigma': value},{'sigma': self.__schema['sigma']}) else ('\x1B[91m','\x1B[0m')
             self.__do_inp_output('   \x1B[92m%-10.10s =\x1B[0m %s%-23s%s' % ('sigma',pre,self.__to_string_(value),post),description,9+len(pre)+len(post))
    def __sizeRegion_inp(self):
        def xml_default( ):
            ## play the crazy subparameter shell game
            dflt = self.__sizeRegion_dflt( self.__globals_( ) )
            if dflt is not None: return dflt
            return float(20)
        if self.__sizeRegion_dflt( self.__globals_( ) ) is not None:
             description = 'User defined local square regional size.'
             value = self.__sizeRegion( self.__globals_( ) )
             (pre,post) = (('','') if value == xml_default( ) else ('\x1B[34m','\x1B[0m')) if self.__validate_({'sizeRegion': value},{'sizeRegion': self.__schema['sizeRegion']}) else ('\x1B[91m','\x1B[0m')
             self.__do_inp_output('   \x1B[92m%-10.10s =\x1B[0m %s%-23s%s' % ('sizeRegion',pre,self.__to_string_(value),post),description,9+len(pre)+len(post))
    def __radius_inp(self):
        def xml_default( ):
            ## play the crazy subparameter shell game
            dflt = self.__radius_dflt( self.__globals_( ) )
            if dflt is not None: return dflt
            return float(1)
        if self.__radius_dflt( self.__globals_( ) ) is not None:
             description = 'User defined radius value for gradient algorithm.'
             value = self.__radius( self.__globals_( ) )
             (pre,post) = (('','') if value == xml_default( ) else ('\x1B[34m','\x1B[0m')) if self.__validate_({'radius': value},{'radius': self.__schema['radius']}) else ('\x1B[91m','\x1B[0m')
             self.__do_inp_output('   \x1B[92m%-10.10s =\x1B[0m %s%-23s%s' % ('radius',pre,self.__to_string_(value),post),description,9+len(pre)+len(post))
    def __ignorPoint_inp(self):
        def xml_default( ):
            ## play the crazy subparameter shell game
            dflt = self.__ignorPoint_dflt( self.__globals_( ) )
            if dflt is not None: return dflt
            return True
        if self.__ignorPoint_dflt( self.__globals_( ) ) is not None:
             description = 'User defined weather ignore the current point regional mmean algorithm.'
             value = self.__ignorPoint( self.__globals_( ) )
             (pre,post) = (('','') if value == xml_default( ) else ('\x1B[34m','\x1B[0m')) if self.__validate_({'ignorPoint': value},{'ignorPoint': self.__schema['ignorPoint']}) else ('\x1B[91m','\x1B[0m')
             self.__do_inp_output('   \x1B[92m%-10.10s =\x1B[0m %s%-23s%s' % ('ignorPoint',pre,self.__to_string_(value),post),description,9+len(pre)+len(post))

    #--------- global default implementation-------------------------------------------
    @static_var('state', __sf__('casa_inp_go_state'))
    def set_global_defaults(self):
        self.set_global_defaults.state['last'] = self
        glb = self.__globals_( )
        if 'method' in glb: del glb['method']
        if 'radius' in glb: del glb['radius']
        if 'sigma' in glb: del glb['sigma']
        if 'sizeRegion' in glb: del glb['sizeRegion']
        if 'ignorPoint' in glb: del glb['ignorPoint']
        if 'binnedvis' in glb: del glb['binnedvis']


    #--------- inp function -----------------------------------------------------------
    def inp(self):
        print("# msuvbinflag -- %s" % self._info_desc_)
        self.term_width, self.term_height = shutil.get_terminal_size(fallback=(80, 24))
        self.__binnedvis_inp( )
        self.__method_inp( )
        self.__sigma_inp( )
        self.__sizeRegion_inp( )
        self.__radius_inp( )
        self.__ignorPoint_inp( )

    #--------- tget function ----------------------------------------------------------
    @static_var('state', __sf__('casa_inp_go_state'))
    def tget(self,savefile=None):
        from casashell.private.stack_manip import find_frame
        from runpy import run_path
        filename = savefile
        if filename is None:
            filename = "msuvbinflag.last" if os.path.isfile("msuvbinflag.last") else "msuvbinflag.saved"
        if os.path.isfile(filename):
            glob = find_frame( )
            newglob = run_path( filename, init_globals={ } )
            for i in newglob:
                glob[i] = newglob[i]
            self.tget.state['last'] = self
        else:
            print("could not find last file: %s\nsetting defaults instead..." % filename)
            self.set_global_defaults( )

    #--------- tput function ----------------------------------------------------------
    def tput(self,outfile=None):
        def noobj(s):
           if s.startswith('<') and s.endswith('>'):
               return "None"
           else:
               return s

        _postfile = outfile if outfile is not None else os.path.realpath('msuvbinflag.last')

        _invocation_parameters = OrderedDict( )
        _invocation_parameters['binnedvis'] = self.__binnedvis( self.__globals_( ) )
        _invocation_parameters['method'] = self.__method( self.__globals_( ) )
        _invocation_parameters['sigma'] = self.__sigma( self.__globals_( ) )
        _invocation_parameters['sizeRegion'] = self.__sizeRegion( self.__globals_( ) )
        _invocation_parameters['radius'] = self.__radius( self.__globals_( ) )
        _invocation_parameters['ignorPoint'] = self.__ignorPoint( self.__globals_( ) )

        try:
            with open(_postfile,'w') as _f:
                for _i in _invocation_parameters:
                    _f.write("%-10s = %s\n" % (_i,noobj(repr(_invocation_parameters[_i]))))
                _f.write("#msuvbinflag( ")
                count = 0
                for _i in _invocation_parameters:
                    _f.write("%s=%s" % (_i,noobj(repr(_invocation_parameters[_i]))))
                    count += 1
                    if count < len(_invocation_parameters): _f.write(",")
                _f.write(" )\n")
        except: return False
        return True

    def __call__( self, binnedvis=None, method=None, sigma=None, sizeRegion=None, radius=None, ignorPoint=None ):
        def noobj(s):
           if s.startswith('<') and s.endswith('>'):
               return "None"
           else:
               return s
        _prefile = os.path.realpath('msuvbinflag.pre')
        _postfile = os.path.realpath('msuvbinflag.last')
        task_result = None
        _arguments = [binnedvis,method,sigma,sizeRegion,radius,ignorPoint]
        _invocation_parameters = OrderedDict( )
        if any(map(lambda x: x is not None,_arguments)):
            # invoke python style
            # set the non sub-parameters that are not None
            local_global = { }
            if binnedvis is not None: local_global['binnedvis'] = binnedvis
            if method is not None: local_global['method'] = method

            # the invocation parameters for the non-subparameters can now be set - this picks up those defaults
            _invocation_parameters['binnedvis'] = self.__binnedvis( local_global )
            _invocation_parameters['method'] = self.__method( local_global )

            # the sub-parameters can then be set. Use the supplied value if not None, else the function, which gets the appropriate default
            _invocation_parameters['sigma'] = self.__sigma( _invocation_parameters ) if sigma is None else sigma
            _invocation_parameters['sizeRegion'] = self.__sizeRegion( _invocation_parameters ) if sizeRegion is None else sizeRegion
            _invocation_parameters['radius'] = self.__radius( _invocation_parameters ) if radius is None else radius
            _invocation_parameters['ignorPoint'] = self.__ignorPoint( _invocation_parameters ) if ignorPoint is None else ignorPoint

        else:
            # invoke with inp/go semantics
            _invocation_parameters['binnedvis'] = self.__binnedvis( self.__globals_( ) )
            _invocation_parameters['method'] = self.__method( self.__globals_( ) )
            _invocation_parameters['sigma'] = self.__sigma( self.__globals_( ) )
            _invocation_parameters['sizeRegion'] = self.__sizeRegion( self.__globals_( ) )
            _invocation_parameters['radius'] = self.__radius( self.__globals_( ) )
            _invocation_parameters['ignorPoint'] = self.__ignorPoint( self.__globals_( ) )
        try:
            with open(_prefile,'w') as _f:
                for _i in _invocation_parameters:
                    _f.write("%-10s = %s\n" % (_i,noobj(repr(_invocation_parameters[_i]))))
                _f.write("#msuvbinflag( ")
                count = 0
                for _i in _invocation_parameters:
                    _f.write("%s=%s" % (_i,noobj(repr(_invocation_parameters[_i]))))
                    count += 1
                    if count < len(_invocation_parameters): _f.write(",")
                _f.write(" )\n")
        except: pass
        try:
            _logging_state_ = None
            assert _pc.validate(_invocation_parameters,self.__schema), create_error_string(_pc.errors)
            _logging_state_ = _start_log( 'msuvbinflag', [ 'binnedvis=' + repr(_pc.document['binnedvis']), 'method=' + repr(_pc.document['method']), 'sigma=' + repr(_pc.document['sigma']), 'sizeRegion=' + repr(_pc.document['sizeRegion']), 'radius=' + repr(_pc.document['radius']), 'ignorPoint=' + repr(_pc.document['ignorPoint']) ] )
            task_result = _msuvbinflag_t( _pc.document['binnedvis'],_pc.document['method'],_pc.document['sigma'],_pc.document['sizeRegion'],_pc.document['radius'],_pc.document['ignorPoint'] )
        except Exception as e:
            from traceback import format_exc
            from casatasks import casalog
            casalog.origin('msuvbinflag')
            casalog.post("Exception Reported: Error in msuvbinflag: %s" % str(e),'SEVERE')
            casalog.post(format_exc( ))
            raise #exception is now raised
            #task_result = False
        finally:
            try:
                os.rename(_prefile,_postfile)
            except: pass
            if _logging_state_:
                task_result = _end_log( _logging_state_, 'msuvbinflag', task_result )

        #Added if _msuvbinflag_t returns False and does not raise an exception.
        if task_result is False:
            raise

        return task_result #Still needed

msuvbinflag = _msuvbinflag( )

